#if !defined( ALGORITHM )
	#define ALGORITHM

#include <iostream>
#include <fstream>
#include <vector>
#include "../shared/classes.cpp"
#include "../shared/leaf_helpers.hpp"
#include "../shared/get_range.hpp"
#include "../shared/alg_file.hpp"
#include "../shared/ml.hpp"
using namespace std;

short unsigned int label3o7( rat r0 );
leaf get_preimage( leaf l, leaf major );
bool leaf_cross( leaf one, leaf two );
void write_tableau( vector<short unsigned> vc_tableau, vector<char> vc_connections, int period );
// unsigned get_min_cols( mpz_class );

#if !defined( SYMB_CONSTS )
	#define SYMB_CONSTS
const short unsigned R1 = 1;
const short unsigned R2 = 2;
const short unsigned R3 = 3;
const short unsigned UC = 4;
const short unsigned BC = 5;
const short unsigned L3 = 6;
const short unsigned L2 = 7;
const short unsigned L1 = 8;
const short unsigned L = 9;
const short unsigned R = 10;
const short unsigned C = 11;
#endif

//------------------------------------------------------------------------------
//	leaf algorithm( leaf mu_p, bool verbose ) - if mu_p is > l_{1/7, 2/7} then this
//			returns an endpoint of the minor leaf of mu_q s.t.
//			s_{3/7, 4/7} \mate mu_q = s_{1/7, 2/7} \mate mu_q. If qml.out has been
//			used to generate the minor leaves of the period of mu_p then the
//			both endpoints of mu_q will be returned.
//				If verbose is true other usefull info is printed such as the
//			tableau, full orbit of mu_q etc.
//
//			This prog returns errors in certain situations:
//			error		reason
//			1, 0, 1, 0		trivial mu_p
//			2, 0, 2, 0		not a valid minor leaf
//			3, 0, 3, 0		not a valid minor leaf ( self intersecting forward image )
//			4, 0, 4, 0		uninterpretable vague word
//			5, 0, 5, 0		mu_q could not be found in the range given by the alg.
//			6, 0, 6, 0		vague and full word differ in length
//------------------------------------------------------------------------------

leaf algorithm(  leaf mu_p, bool verbose = true, short unsigned min_cols = 2,
	bool flip = false  ){

	char c;
	short* word;
	int period = 0, j, cIndex = 0, i = 0, k;
	mpz_class n, d, z1, z2;
	rat r0( 0, 0 ), r1, q, one( 1, 1 ), half( 1, 2 );
	leaf mu_q, temp, temp1, temp2, range1, range2, range3, range4, EOFL;

//----------------------------------------------------------------------------//
//----- Check of mu_p --------------------------------------------------------//
//----------------------------------------------------------------------------//

	// 	Check this mu_p is a minor leaf ---------------------------------------
	if( mu_p.get_a() == mu_p.get_b() ){
		if( verbose )
			cout << "\nThat is not a minor leaf (it is a trivial leaf).\n";
		EOFL.set_leaf( 1, 0, 1, 0 );
		return EOFL;
	}

	// Get period = p.
	if ( mu_p.get_a().get_denom() % 2 != 0 && mu_p.get_b().get_denom() % 2 != 0 ) {
		while( ( ++period <= mu_p.get_a().get_denom() ) &&
			( ( j = ( to_pow( 2, period ) - 1 ) ) % mu_p.get_a().get_denom() != 0 ) &&
			( j % mu_p.get_b().get_denom() != 0 ) ) {
			#ifdef PERIOD_DEBUG
				cout << "j = " << j << endl;
				cout << "j % a.denom() = " << j % mu_p.get_a().get_denom() << endl;
				cout << "j % b.denom() = " << j % mu_p.get_b().get_denom() << endl;
			#endif
			}
		if ( period > mu_p.get_a().get_denom() ) {
			if( verbose )
				cout << "\nThat is not a minor leaf (endpoints don't share period).\n";
			EOFL.set_leaf( 2, 0, 2, 0 );
	        return EOFL;
		}
	} else {
		if( verbose )
			cout << "\nThat is not a minor leaf (endpoints not periodic).\n";
		EOFL.set_leaf( 2, 0, 2, 0 );
        return EOFL;
	}

	// Check that no leaf in mu_p's forward image is shorter than mu_p
	temp = mu_p;
	r0 = mu_p.get_length();
	temp.map_forward();
	if( r0 > temp.get_length() ){
        if( verbose )
            cout << "\nThat is not a valid minor leaf.\n";
        EOFL.set_leaf( 2, 0, 2, 0 );
        return EOFL;
    }
	for( ; i < ( period - 2 ); ++i ){
		temp.map_forward();
		if( r0 > temp.get_length() ){
    	    if( verbose )
    	        cout << "\nThat is not a valid minor leaf.\n";
    	    EOFL.set_leaf( 2, 0, 2, 0 );
    	    return EOFL;
    	}
	}


//----------------------------------------------------------------------------//
//----- Calculation of orbits and finding which leaf -T approximates ---------//
//----------------------------------------------------------------------------//

	// Calc orbit of mu_p ------------------------------------------------------
	leaf mu_pOrbit[period];
	mu_pOrbit[0] = mu_p;
	for( j = 1; j < period; ++j ){
		mu_pOrbit[j] = mu_pOrbit[ j - 1 ];
		mu_pOrbit[j].map_forward();
	}

	// Check that mu_pOrbit doesn't cross itself
	for( j = 0; j < period; ++j ){
		for( k = j + 1; k < period; ++k ){
			if( leaf_cross( mu_pOrbit[j], mu_pOrbit[k] ) ){
				if( verbose ){
					cout << "\nThat is not a valid minor leaf (its orbit is ";
					cout << "self intersecting).\n";
				}
				EOFL.set_leaf( 3, 0, 3, 0 );
				return EOFL;
			}
 		}
	}

	if( verbose ){
		cout << "\nThat is a minor leaf of period " << period << ".\n\n";
		cout << "The forward orbit of the minor leaf, " << mu_p << " is:\n";
		for( j = 1; j < period; ++j ){
			cout << mu_pOrbit[j] << endl;
		}
		cout << endl;
	}

	// Calc pre-images of T which approx \mu_p until enought letters of the
	// vague word has been found
		// Find the pre-image of mu_p which is best approximated by -T
		// (or T), take a note of it (save cIndex so that -T (or T)
		// approximates mupOrbit[cIndex] )
	r0.set_rat( 1, 2 );
	#ifndef _T_
		tri T( 1, 7, 2, 7, 4, 7 ), Tm( 1, 14, 9, 14, 11, 14 ), t = Tm;	// T is T (?! ),
	#else
		tri T( 1, 7, 2, 7, 4, 7 ), Tm( 1, 14, 9, 14, 11, 14 ), t = T;	// T is T (?! ),
	#endif
										// Tm is -T, t is the current
										// approximation triangle

	r0.set_rat( 1, 1 );
	for( j = 0; j < period; ++j ){
		if( ( r1 = dist( mu_pOrbit[j], t ) ) < r0 ){
			r0 = r1;
			cIndex = j;
		}
	}
	if( verbose ){
		cout << "Triangle ";
		#ifndef _T_
			cout << "-T";
		#else
			cout << "T";
		#endif
		cout << " is approximating the " << cIndex
			 << "th forward image (" << mu_pOrbit[ cIndex ] << ") of mu_p.\n";
	}

	// Get vertex of t which is closest to mu_p
	{
		rat l1, l2, l3;
		rat d1, d2, d3;

		l1 = t.get_a();
		l2 = t.get_b();
		l3 = t.get_c();

		// decide which endpoint of mu_pOrbit[cIndex] we want to approx.
		if( !flip ){
			r1 = mu_pOrbit[ cIndex ].get_a();
		}else{
			r1 = mu_pOrbit[ cIndex ].get_b();
		}

		d1 = dist( l1, r1 );
		d2 = dist( l2, r1 );
		d3 = dist( l3, r1 );

		if( d1 < d2 ){
			if( d1 < d3 ){
				r0 = l1;
			}else{
				r0 = l3;
			}
		}else{
			if( d2 < d3 ){
				r0 = l2;
			}else{
				r0 = l3;
			}
		}
	}
// Uncomment to have -T always start approximating a pre-image of the minor leaf
/*	temp.set_leaf( t.get_a(), t.get_b() );
	cIndex = period - 1; */



//----------------------------------------------------------------------------//
//----- Convert T, Tm, r0, mu_p and it's orbit into the inside labelling -----//
//----------------------------------------------------------------------------//

	T.set_tri( 3, 7, 5, 7, 6, 7 );
	Tm.set_tri( 13, 14, 3, 14, 5, 14 );

	r0 = one - r0;

	for( j = 0; j < period; ++j ){
		mu_pOrbit[j].set_a( one - mu_pOrbit[j].get_a() );
		mu_pOrbit[j].set_b( one - mu_pOrbit[j].get_b() );
	}
	mu_p = mu_pOrbit[0];
	// Now all labels/rationals use the inside labelling




//----------------------------------------------------------------------------//
//----- Calculation of first row of tableau ----------------------------------//
//----------------------------------------------------------------------------//

	bool faltered = false;
	int rows = 1, columns = period + 1 + cIndex;

	// Calculate how many columns are required for finding mu_q
	j = min_cols * period;
	//j = get_min_cols( period ); //mu_p.get_a().get_denom() );
	while( columns < j )
		columns +=  period;
	word = new short [ columns ];

	vector< vector<rat> > 				vvrat_tableau( rows, vector<rat>( columns, one ) );
	vector< vector<short unsigned> > 	vvsui_tableau( rows, vector<short unsigned>( columns, 0 ) );
	vector< vector<char> > 				vvc_connections( rows, vector<char>( columns, '?' ) );
	vector< char > 						vc_vague_word( columns, '?' );

	// Put one endpoint of the leaf approximated by T into vvrat_tableau. If flip
	// is true approx \mu_q with larger endpoint of \mu_p, else smaller.
	vvrat_tableau[0][0] = r0;

	//    Get pre-image approximations of mu_p
	for( j = 1; j < columns; ++j ){
		if( cIndex == 0 )
			cIndex = period - 1;
		else
			cIndex--;

		r0.set_rat( 1, 2 );
		r1.set_rat( vvrat_tableau[0][ j - 1 ].get_num(), vvrat_tableau[0][ j - 1 ].get_denom()  <<  1 );
		r0 = r1 + r0;
		if( dist( r1, mu_pOrbit[ cIndex ] ) < dist( r0, mu_pOrbit[ cIndex ] ) ){
			vvrat_tableau[0][j] = r1;
		}else{
			vvrat_tableau[0][j] = r0;
		}
	}
#ifdef DEBUG
	cout << "\n\n";
	for( j = columns - 1; j >= 0; --j )
		cout << vvrat_tableau[0][j] << ' ';
	cout << "\n\n";
#endif
	//	Label points
	for( j = 0; j < columns; ++j ){
		vvsui_tableau[0][j] = label3o7( vvrat_tableau[0][j] );
	}




//----------------------------------------------------------------------------//
//----- Main loop ------------------------------------------------------------//
//----------------------------------------------------------------------------//

	while( faltered == false ){
		faltered = true;
		// Label connections below rows and update leaf_orb

		// Start at the second letter
		for( j = 0; vvsui_tableau[ rows - 1 ][j] == 0; ++j );
		//j++;

		for( ; j < columns; ++j ){
			range3.set_leaf( 0, 0, 0, 0 );
			range4.set_leaf( 0, 0, 0, 0 );

			// prepare the word
			for( k = 0; k < j + 1; ++k )
				word[ columns - 1 - k ] = vvsui_tableau[ rows - 1 ][k];
			for( k = j + 1; k < columns; ++k )
				word[ columns - 1 - k ] = 0;

			switch( get_word_label( word, columns ) ){
				case 2:	// -T connection found
					// empty row of vvrat_tableau
					for( k = 0; k < columns; ++k ){
						vvrat_tableau[ rows - 1 ][k].set_rat( 0, 0 );
					}

					// put new rotation in upper row
					if( ( vvsui_tableau[ rows - 1 ][j] == C ) ||
							( vvsui_tableau[ rows - 1 ][j] == UC ) ||
							( vvsui_tableau[ rows - 1 ][j] == BC ) ||
							( vvsui_tableau[ rows - 1 ][j] == R3 ) )
						vvrat_tableau[ rows - 1 ][j].set_rat( 3, 14 );
					else if( ( vvsui_tableau[ rows - 1 ][j] == R1 ) ||
							( vvsui_tableau[ rows - 1 ][j] == R2 ) )
						vvrat_tableau[ rows - 1 ][j].set_rat( 13, 14 );
					else if( ( vvsui_tableau[ rows - 1 ][j] == L2 ) ||
							( vvsui_tableau[ rows - 1 ][j] == L3 ) )
						vvrat_tableau[ rows - 1 ][j].set_rat( 5, 14 );

					// label connection
		  			vvc_connections[ rows - 1 ][j] = 'R';

					// break loop
					j = columns;

					break;
				case 1:	// T connection found
					// label connection
		  			vvc_connections[ rows - 1 ][j] = 'L';
					break;
				default:
					break;
			}
		}

		// Update vague word
		for( k = 0; k < columns; ++k ){
			if( vc_vague_word[k] == '?' ){
				if( vvc_connections[ rows - 1 ][k] !=  '?' )
					vc_vague_word[k] = vvc_connections[ rows - 1 ][k];
			}
		}
		k--;
		while( vc_vague_word[k] !=  'R' && k >=  0 ) k--;
		for( ; k >=  0; k-- ){
			if( vc_vague_word[k] == '?' ){
				if( vvsui_tableau[ rows - 1 ][k] == BC )
					vc_vague_word[k] = 'B';
				else if( vvsui_tableau[ rows - 1 ][k] == UC )
					vc_vague_word[k] = 'U';
			}
		}

		if( verbose )
			write_tableau( vvsui_tableau[ rows - 1 ], vvc_connections[ rows - 1 ], period );

		// Add new row if a -T connection was found and there are more columns
		// to label
		if( vc_vague_word[ columns - 1 ] == '?' ){
			// Add row
			vector<rat> 			vrat_t( columns );
			vector<short unsigned> 	vsui_t( columns, 0 );
			vector<char>			vc_c( columns, '?' );
			for( k = 0; k < vrat_t.size(); ++k ) vrat_t[k].set_rat( 0, 0 );

			++rows;
			vvrat_tableau.push_back( vrat_t );
			vvsui_tableau.push_back( vsui_t );
			vvc_connections.push_back( vc_c );

			// Calculate first entry on new row
			for( j = columns - 1; j >=  0; j-- ){
				if( vvc_connections[ rows - 2 ][j] == 'R' ){
					vvrat_tableau[ rows - 1 ][j].set_rat( 3, 14 );
					vvsui_tableau[ rows - 1 ][j] = C;
					j = -1;
					faltered = false;
				}
			}

			// Find the pre-images of the new leaves
			for( j = 0; j < columns - 1; ++j ){
				// If there's something to take a pre-image of
				if( vvrat_tableau[ rows - 1 ][j].get_denom() !=  0 ){
					// Get the possible pre-images
					temp.set_leaf( vvrat_tableau[rows-2][j], vvrat_tableau[ rows - 1 ][j] );
					temp1 = get_preimage( temp, mu_pOrbit[ period - 1 ] );
					temp2 = temp1 + half;

					// Find the correct region for the higher endpoint to be in
					k = 0;
					// Discard any letters after the R connection
					while( vvc_connections[ rows - 2 ][k] !=  'R' )
						word[ k++ ] = 0;
					for( ; k < j + 2; ++k )
						word[ columns - 1 - k ] = vvsui_tableau[ rows - 2 ][k];
					for( k = j + 2; k < columns; ++k )
						word[ columns - 1 - k ] = 0;
					range1 = get_aero_range( word, columns );
					range2 = get_aero_range( word, columns, false );

					// Is the first endpoint of range1 in the correct region?
					if( ( rat_in_CLOSED( temp1.get_a(), range1 ) ) || ( rat_in_CLOSED( temp1.get_a(), range2 ) ) ){
						vvrat_tableau[ rows - 2 ][ j + 1 ] = temp1.get_a();
						vvrat_tableau[ rows - 1 ][ j + 1 ] = temp1.get_b();
					}else if( ( rat_in_CLOSED( temp2.get_a(), range1 ) ) || ( rat_in_CLOSED( temp2.get_a(), range2 ) ) ){
						vvrat_tableau[ rows - 2 ][ j + 1 ] = temp2.get_a();
						vvrat_tableau[ rows - 1 ][ j + 1 ] = temp2.get_b();
					}
				}
			}

			// Label endpoints of the pullbacks
			for( j = 0; j < columns; ++j ){
				if( vvsui_tableau[ rows - 1 ][j] == 0 )
					vvsui_tableau[ rows - 1 ][j] = label3o7( vvrat_tableau[ rows - 1 ][j] );
			}
		}
	}




//----------------------------------------------------------------------------//
//----- Process vague word ---------------------------------------------------//
//----------------------------------------------------------------------------//

	// Add any missing Us or Bs to Vague word
	for( j = 0; j < columns; ++j ){
		if( vc_vague_word[j] == '?' ){
			k = 1;
			while( vvsui_tableau[rows - k][j] == 0 ) ++k;
			if( vvsui_tableau[ rows - k ][j] == BC )
				vc_vague_word[j] = 'B';
			else if( vvsui_tableau[ rows - k ][j] == UC )
				vc_vague_word[j] = 'U';
		}
	}

	// Output vague word
	if( verbose ){
		cout << "\nThe vague periodic word is\n ";
		for( k = 0; k < period; ++k ){
			if( vc_vague_word[ columns - k - 1 ] !=  'U' && vc_vague_word[ columns - k - 1 ] !=  'B' )
				cout << vc_vague_word[ columns - k - 1 ] << "   ";
			else
				cout << vc_vague_word[ columns - k - 1 ] << "C  ";
		}
	}

	// Translate vague word into full word -------------------------------------
		// Put vc_vague word into string
	string vague;
	// Get at most period chars of vc_vague_word into vague
	for( j = 0; j < period; ++j )
		vague += vc_vague_word[ columns - 1 - j ];
	// add one extra char to help determine the final periodic letter
	vague += vague.at(0);

	// expand Bs and Us to BCs and UCs
	for( j = 0; j < vague.length(); ++j ){
		if( vague.at(j) == 'B' || vague.at(j) == 'U' )
			vague.insert( ++j, 1, 'C' );
	}

	columns = trans_v1( vague, word ); // columns now holds the length of word
	// discard the letter that was added to help determine the final periodic
	// letter
	word[ period ] = 0;
	--columns;

	// Check that the full word that's been returned is valid
	if( !chk_word1( word, columns ) ){
		if( verbose )
			cout << "\n\nThat vague word does not translate into a valid full word.\n";
		EOFL.set_leaf( 4, 0, 4, 0 );
		return EOFL;
	}
	if( verbose ){
		cout << "\n\nThe full periodic word is\n ";
		for( k = 0; k < period; ++k ){
			letter_out( word[k] );
			if( word[k] !=  L && word[k] !=  R && word[k] !=  C ){
				cout << "  ";
			}else{
				cout << "   ";
			}
		}
		cout << endl;
	}
	// Check that the full word is the same length as the vague word
	j = k = 0;
	#ifdef VAGUE_DEBUG
		cout << "\nvague.length() = " << vague.length() << endl;
	#endif
	for( i = 0; i < vague.length(); ++i ){
		if( vague.at(i) !=  '?' && vague.at(i) != 'C' )
			++j;
		/* if( word[i] !=  0 ){
			#ifdef VAGUE_DEBUG
				cout << "i = " << i << ", k = " << k << ": letter is \"";
				letter_out( word[ k ] );
				cout << '\"' << endl;
			#endif
			++k;
		} Now user number stored in 'columns' from trans_v1() call */
	}
	if( j != columns + 1 ){		// as vague still has extra final letter for helping to
							// determine the last periodic letter of word.
		if( verbose ){
			cout << "which is not the same length as the vague word. Hmmm...\n";
			cout << "Vague word length = " << j << endl;
			cout << "Rabbit word length = " << columns << endl;
		}
		EOFL.set_leaf( 6, 0, 6, 0 );
		return EOFL;
	}




//----------------------------------------------------------------------------//
// ----- Find q from word ----------------------------------------------------//
//----------------------------------------------------------------------------//

	range1 = get_rabbit_range( word, columns );
	if( verbose )
		cout << "\nRange given by periodic word is\n" << range1 << " (inside labelling)" << endl;

	mpz_ui_pow_ui( n.get_mpz_t(), 2, period );
	--n;	// n = 2 ^ period - 1
	q.set_rat( 1, n );

	while( !rat_in_CLOSED( q, range1 ) && q.get_num() < q.get_denom() )
		q.set_num( q.get_num() + 1 );
	if( q.get_num() !=  q.get_denom() ){
		q.simplify();
		q.set_num( q.get_denom() - q.get_num() );	// use outside labelling
		if( verbose )
			cout << "\nq = " << q;
	}else{
		if( verbose ){
			cout << "\nCould not find n s.t. n/" << q.get_denom() << " is in the ";
			cout << "reqired range of " << range1 << endl;
		}
		EOFL.set_leaf( 5, 0, 5, 0 );
		return EOFL;
	}

	// Calc mu_q ---------------------------------------------------------------
	mu_q.set_a( q );
	mu_q.set_b( find_minor( mu_q.get_a(), period ) );

	if( verbose )
		cout << "\n\n\\mu_q = " << mu_q;

	// Return
	if( verbose )
		cout << endl;
	return mu_q;
}

//------------------------------------------------------------------------------
// write_tableau( short unsigned* tableau, char* connections, int rows,
// 		int columns, short prow = -1 )
//			- writes a line of the tableau to standard out ( for convenience and
//			debugging! )
//
//			NOTE: if this is called and prow is greater than the number or
//					rows in tableau or connections then there'll be random data
//					given out.
//------------------------------------------------------------------------------

void write_tableau( vector<short unsigned> vsui_tableau, vector<char> vc_connections, int period )
{
	int j;
	int columns = vsui_tableau.size();

	// print out letters
	for( j = ( columns - 1 ); j >= 0; j-- ){
		letter_out( vsui_tableau[j] );
		if( vsui_tableau[j] == C )
			cout << ' ';
		if( ( columns - j ) % period !=  0 ){
			cout << "  ";
		}else{
			if( vsui_tableau[j] !=  0 )
				cout << " | ";
		}
	}
	cout << endl;

	// print out connections
	for( j = ( columns - 1 ); j >= 0; j-- ){
		if( ( columns - j ) % period !=  0 ){
			if( vc_connections[j] == '?' ){
				if( vsui_tableau[j] !=  0 )
					cout << "|   ";
			}else{
				if( vsui_tableau[j] !=  0 )
				cout << '|' << vc_connections[j] << "  ";
			}
		}else{
			if( vc_connections[j] == '?' ){
				if( vsui_tableau[j] !=  0 )
					cout << "|  | ";
			}else{
				if( vsui_tableau[j] !=  0 )
					cout << '|' << vc_connections[j] << " | ";
			}
		}
	}
	cout << flush << endl;
}

//------------------------------------------------------------------------------
// label3o7( rat ) - returns the label of rat in dynamics of 3/7
//                 All intervals are closed below and open above
//                 i.e. L1 = [3/7, 4/7 )
//------------------------------------------------------------------------------

short unsigned int label3o7( rat r0 )
{
	// Ignore null leaves
	if( r0.get_denom() == 0 )
		return 0;

	// Register 3/14 and 5/7 as C
	rat r1( 5, 7 ), r2( 3, 14 );
	//if( ( r0 == r1 ) || ( r0 == r2 ) )
	//	return C;

	r1.set_rat(  1, 14 );
	r2.set_rat( 13, 14 );
	if( r0 < r1 || r0 >=  r2 )
		return R1;
	r1.set_rat( 1, 7 );
	r2.set_rat( 6, 7 );
	if( r0 < r1 || r0 >=  r2 )
		return R2;
	r1.set_rat(  3, 14 );
	r2.set_rat( 11, 14 );
	if( r0 < r1 || r0 >=  r2 )
		return R3;
	r1.set_rat( 3, 14 );
	r2.set_rat( 2,  7 );
	if( r0 < r2 && r0 >=  r1 )
		return UC;
	r1.set_rat(  5,  7 );
	r2.set_rat( 11, 14 );
	if( r0 < r2 && r0 >=  r1 )
		return BC;
	r1.set_rat( 5, 14 );
	r2.set_rat( 9, 14 );
	if( r0 < r1 || r0 >=  r2 )
		return L3;
	r1.set_rat( 3, 7 );
	r2.set_rat( 4, 7 );
	if( r0 < r1 || r0 >=  r2 )
		return L2;

	return L1;
}

//------------------------------------------------------------------------------
//	get_min_cols( ull d ) - returns the minimum length for a word in L_{1/7} to
//						specify a periodic point with denominator d
//------------------------------------------------------------------------------

/*mpz_class get_min_cols( mpz_class d )
{
	/*unsigned l = 1;
	rat range( 2, 7 );

    while( range.get_denom() < d ){
        range.set_denom( range.get_denom() * 2 );
        l++;
    }/

	return 2 * d;
}*/
#endif
